from scipy.stats import norm
from ..centile_bands import centile_band_for_centile
from ..global_functions import sds_for_centile
from ..constants.reference_constants import HEIGHT, WEIGHT, HEAD_CIRCUMFERENCE, BMI
from ..constants.reference_constants import THREE_PERCENT_CENTILES, FIVE_PERCENT_CENTILES, EIGHTY_FIVE_PERCENT_CENTILES, COLE_TWO_THIRDS_SDS_NINE_CENTILES

measurements = [HEIGHT, WEIGHT, HEAD_CIRCUMFERENCE, BMI]
measurement_texts = ["height", "weight", "head circumference", "body mass index"]

on_cole_centiles = [
            {
                "centile": 0.4, 
                "text": "0.4th"
            },
            {
                "centile": 2.0, 
                "text": "2nd"
            },
            {
                "centile": 9.0, 
                "text": "9th"
            },
            {
                "centile": 25.0, 
                "text": "25th"
            },
            {
                "centile": 50.0, 
                "text": "50th"
            },
            {
                "centile": 75.0, 
                "text": "75th"
            },
            {
                "centile": 91.0, 
                "text": "91st"
            },
            {
                "centile": 98.0, 
                "text": "98th"
            },
            {
                "centile": 99.6, 
                "text": "99.6th"
            }
        ]
between_cole_centiles = [
            {
                "centile": 1.0, 
                "text": "between the 0.4th and 2nd"
            },
            {
                "centile": 5.0, 
                "text": "between the 2nd and 9th"
            },
            {
                "centile": 17.0, 
                "text": "between the 9th and 25th"
            },
            {
                "centile": 37.0, 
                "text": "between the 25th and 50th"
            },
            {
                "centile": 62.0, 
                "text": "between the 50th and 75th"
            },
            {
                "centile": 83.0, 
                "text": "between the 75th and 91st"
            },
            {
                "centile": 94.5, 
                "text": "between the 91st and 98th"
            },
            {
                "centile": 99.3, 
                "text": "between the 98th and 99.6th"
            }
        ]

beyond_cole_thresholds = [
            {
                "centile": 99.8, 
                "text": "above the normal range"
            },
            {
                "centile": 0.1, 
                "text": "below the normal range"
            }
        ]

# on three-percent-centiles (3rd, 5th, 10th, 25th, 50th, 75th, 90th, 95th, 97th)
on_three_percent_centiles = [
            {
                "centile": 3.0, 
                "text": "3rd"
            },
            {
                "centile": 5.0, 
                "text": "5th"
            },
            {
                "centile": 10.0, 
                "text": "10th"
            },
            {
                "centile": 25.0, 
                "text": "25th"
            },
            {
                "centile": 50.0, 
                "text": "50th"
            },
            {
                "centile": 75.0, 
                "text": "75th"
            },
            {
                "centile": 90.0, 
                "text": "90th"
            },
            {
                "centile": 95.0, 
                "text": "95th"
            },
            {
                "centile": 97.0, 
                "text": "97th"
            }
        ]

# between three-percent-centiles (7.0th, 17.5th, 37.5th, 62.5th, 82.5th, 93rd)
between_three_percent_centiles = [
            {
                "centile": 7.0, 
                "text": "between the 5th and 10th"
            },
            {
                "centile": 17.5, 
                "text": "between the 10th and 25th"
            },
            {
                "centile": 37.5, 
                "text": "between the 25th and 50th"
            },
            {
                "centile": 62.5, 
                "text": "between the 50th and 75th"
            },
            {
                "centile": 82.5, 
                "text": "between the 75th and 90th"
            },
            {
                "centile": 93, 
                "text": "between the 90th and 95th"
            },
        ]

# beyond three-percent-centiles (1st, 97.5th, 99.9th)
beyond_three_percent_thresholds = [
            {
                "centile": 1.0, 
                "text": "below the normal range"
            },
            {
                "centile": 98, 
                "text": "above the normal range"
            },
            {
                "centile": 99.9,
                "text": "above the normal range"
            }
        ]

# on five-percent-centiles (5th, 10th, 25th, 50th, 75th, 90th, 95th)
on_five_percent_centiles = [
            {
                "centile": 5.0,
                "text": "5th",
            },
            {
                "centile": 10.0,
                "text": "10th",
            },
            {
                "centile": 25.0,
                "text": "25th",
            },
            {
                "centile": 50.0,
                "text": "50th",
            },
            {
                "centile": 75.0,
                "text": "75th",
            },
            {
                "centile": 90.0,
                "text": "90th",
            },
            {
                "centile": 95.0,
                "text": "95th",
            }
        ]

# between five-percent-centiles (8th, 17.5th, 37.5th, 62.5th, 82.5th, 92.5th)
between_five_percent_centiles = [
            {
                "centile": 7,
                "text": "between the 5th and 10th",
            },
            {
                "centile": 17.5,
                "text": "between the 10th and 25th",
            },
            {
                "centile": 37.5,
                "text": "between the 25th and 50th",
            },
            {
                "centile": 62.5,
                "text": "between the 50th and 75th",
            },
            {
                "centile": 82.5,
                "text": "between the 75th and 90th",
            },
            {
                "centile": 93,
                "text": "between the 90th and 95th",
            }
        ]

# beyond five-percent-centiles (1st, 97.5th, 99.9th)
beyond_five_percent_thresholds = [
            {
                "centile": 1.0,
                "text": "below the normal range",
            },
            {
                "centile": 97.5,
                "text": "above the normal range",
            },
            {
                "centile": 99.9,
                "text": "above the normal range",
            }
        ]

# on eighty-five-percent-centiles (15th, 50th, 85th)
on_eighty_five_percent_centiles_bmi = [
            {
                "centile": 85.0,
                "text": "85th",
            },
            {
                "centile": 98.0,
                "text": "98th",
            },
            {
                "centile": 99.9,
                "text": "99.9th",
            },
            {
                "centile": 99.99,
                "text": "99.99th"
            }
        ]

# between eighty-five-percent-centiles (32.5th, 67.5th)
between_eighty_five_percent_centiles = [
            {
                "centile": 15.0,
                "text": "between the 10th and 25th",
            },
            {
                "centile": 32.5,
                "text": "between the 25th and 50th",
            },
            {
                "centile": 67.5,
                "text": "between the 50th and 75th",
            },
            {
                "centile": 80,
                "text": "between the 75th and 85th",
            }
        ]

# beyond eighty-five-percent-centiles (7.5th, 92.5th)
beyond_eighty_five_percent_thresholds = [
            {
                "centile": 7.5,
                "text": "below the normal range",
            },
            {
                "centile": 92.5,
                "text": "above the normal range",
            }
]

def test_centile_band_for_centile_cole():

    # on cole centiles (2.0, 9.0, 25.0, 50.0, 75.0, 91.0, 98.0, 99.6)
    for measurement in measurements:
        for centile in on_cole_centiles:
            sds = sds_for_centile(centile['centile'])
            assert centile_band_for_centile(sds=sds, measurement_method=measurement, centile_format=COLE_TWO_THIRDS_SDS_NINE_CENTILES) == f"This {measurement_texts[measurements.index(measurement)]} measurement is on or near the {centile['text']} centile."
        
        
    #  between cole centiles (0.4th, 2nd, 9th, 25th, 50th, 75th, 91st, 98th, 99.6th)
    for measurement in measurements:
        for centile in between_cole_centiles:
            sds = sds_for_centile(centile['centile'])
            assert centile_band_for_centile(sds=sds, measurement_method=measurement, centile_format=COLE_TWO_THIRDS_SDS_NINE_CENTILES) == f"This {measurement_texts[measurements.index(measurement)]} measurement is {centile['text']} centiles."

    #  above 99.6th centile or below 0.4th centile
    for measurement in measurements:
        for centile in beyond_cole_thresholds:
            sds = sds_for_centile(centile['centile'])
            assert centile_band_for_centile(sds=sds, measurement_method=measurement, centile_format=COLE_TWO_THIRDS_SDS_NINE_CENTILES) == f"This {measurement_texts[measurements.index(measurement)]} measurement is {centile['text']}."

def test_centile_band_for_centile_three_percent():
    # on three-percent-centiles (3rd, 5th, 10th, 25th, 50th, 75th, 90th, 95th, 97th)
    for measurement in measurements:
        for centile in on_three_percent_centiles:
            sds = sds_for_centile(centile['centile'])
            assert centile_band_for_centile(sds=sds, measurement_method=measurement, centile_format=THREE_PERCENT_CENTILES) == f"This {measurement_texts[measurements.index(measurement)]} measurement is on or near the {centile['text']} centile."
        
    # between three-percent-centiles (4th, 7.5th, 17.5th, 37.5th, 62.5th, 82.5th, 92.5th, 96.5th)
    for measurement in measurements:
        for centile in between_three_percent_centiles:
            sds = sds_for_centile(centile['centile'])
            assert centile_band_for_centile(sds=sds, measurement_method=measurement, centile_format=THREE_PERCENT_CENTILES) == f"This {measurement_texts[measurements.index(measurement)]} measurement is {centile['text']} centiles."

    # beyond three-percent-centiles (1st, 97.5th, 99.9th)
    for measurement in measurements:
        for centile in beyond_three_percent_thresholds:
            sds = sds_for_centile(centile['centile'])
            assert centile_band_for_centile(sds=sds, measurement_method=measurement, centile_format=THREE_PERCENT_CENTILES) == f"This {measurement_texts[measurements.index(measurement)]} measurement is {centile['text']}."

def test_centile_band_for_centile_five_percent():
    # on five-percent-centiles (5th, 10th, 25th, 50th, 75th, 90th, 95th)
    for measurement in measurements:
        for centile in on_five_percent_centiles:
            sds = sds_for_centile(centile['centile'])
            assert centile_band_for_centile(sds=sds, measurement_method=measurement, centile_format=FIVE_PERCENT_CENTILES) == f"This {measurement_texts[measurements.index(measurement)]} measurement is on or near the {centile['text']} centile."
        
    # between five-percent-centiles (7.5th, 17.5th, 37.5th, 62.5th, 82.5th, 92.5th)
    for measurement in measurements:
        for centile in between_five_percent_centiles:
            sds = sds_for_centile(centile['centile'])
            assert centile_band_for_centile(sds=sds, measurement_method=measurement, centile_format=FIVE_PERCENT_CENTILES) == f"This {measurement_texts[measurements.index(measurement)]} measurement is {centile['text']} centiles."

    # beyond five-percent-centiles (1st, 97.5th, 99.9th)
    for measurement in measurements:
        for centile in beyond_five_percent_thresholds:
            sds = sds_for_centile(centile['centile'])
            assert centile_band_for_centile(sds=sds, measurement_method=measurement, centile_format=FIVE_PERCENT_CENTILES) == f"This {measurement_texts[measurements.index(measurement)]} measurement is {centile['text']}."

def test_centile_band_for_centile_eighty_five_percent():
        
    # between eighty-five-percent-centiles (32.5th, 67.5th)
    for measurement in measurements:
        for centile in between_eighty_five_percent_centiles:
            sds = sds_for_centile(centile['centile'])
            assert centile_band_for_centile(sds=sds, measurement_method=measurement, centile_format=EIGHTY_FIVE_PERCENT_CENTILES) == f"This {measurement_texts[measurements.index(measurement)]} measurement is {centile['text']} centiles."

    #on eighty-five-percent-centiles (85th, 98th, 99.9th, 99.99th)
    for centile in on_eighty_five_percent_centiles_bmi:
        sds = sds_for_centile(centile['centile'])
        assert centile_band_for_centile(sds=sds, measurement_method="bmi", centile_format=EIGHTY_FIVE_PERCENT_CENTILES) == f"This body mass index measurement is on or near the {centile['text']} centile."

def test_advisory_thresholds():
    for measurement in measurements:
        for centile in beyond_eighty_five_percent_thresholds:
            sds = 5 # above 4 and less than 8
            if measurement == BMI:
                sds = 10
            assert centile_band_for_centile(sds=sds, measurement_method=measurement, centile_format=COLE_TWO_THIRDS_SDS_NINE_CENTILES) == f"This {measurement_texts[measurements.index(measurement)]} measurement is well outside the normal range. Please check its accuracy."


from rcpchgrowth.constants import CDC, WHO

def comment_prematurity_correction(
    chronological_decimal_age: float,
    corrected_decimal_age: float,
    gestation_weeks: int,
    gestation_days: int,
    reference: str):
    """
    Returns interpretations on age correction as a string
    """

    if chronological_decimal_age == corrected_decimal_age:
        # no adjustment has been made so the child was born at 40 weeks or this child is being plotted on the CDC reference, where no adjustment is made for 37-42 weeks or beyond the age of 2 years
            lay_corrected_decimal_age_comment = "Your child was born on their due date."
            clinician_corrected_decimal_age_comment = "Born at term. No correction has been made for gestation."
            lay_chronological_decimal_age_comment = "Your child was born on their due date."
            clinician_chronological_decimal_age_comment = "Born Term. No correction has been made for gestation."
            # These fields should only apply to CDC reference, since UK-WHO corrects for all gestations (and therefore corrected_decimal_age will never be equal to chronological_decimal_age if gestation_weeks is not 40)
            if gestation_weeks < 42 and (reference == CDC or reference == WHO):
                if gestation_weeks < 37:
                    lay_chronological_decimal_age_comment = f"Your child was born at {gestation_weeks}+{gestation_days} weeks gestation. No correction is made for this beyond 2 years of age."
                    clinician_chronological_decimal_age_comment =f"Born preterm at {gestation_weeks}+{gestation_days} weeks gestation. No correction is made for this beyond 2 years of age."
                    lay_corrected_decimal_age_comment = f"Your child was born at {gestation_weeks}+{gestation_days} weeks gestation. No correction is made for this beyond 2 years of age."
                    clinician_corrected_decimal_age_comment = f"Born preterm at {gestation_weeks}+{gestation_days} weeks gestation. No correction is made for this beyond 2 years of age."
                else:
                    lay_chronological_decimal_age_comment = f"Your child was born at {gestation_weeks}+{gestation_days} weeks gestation. This is considered term and no correction for gestation has been made."
                    clinician_chronological_decimal_age_comment = f"Born at term ({gestation_weeks}+{gestation_days} weeks gestation). No correction for gestation has been made."
                    lay_corrected_decimal_age_comment = f"Your child was born at {gestation_weeks}+{gestation_days} weeks gestation. This is considered term and no correction for gestation has been made."
                    clinician_corrected_decimal_age_comment = f"Born at term ({gestation_weeks}+{gestation_days} weeks gestation). No correction for gestation has been made."
                    if gestation_weeks == 40 and gestation_days == 0:
                        lay_chronological_decimal_age_comment = "Your child was born at term. No correction for gestation has been made."
                        clinician_chronological_decimal_age_comment = "Born at term. No correction for gestation has been made."
                        lay_corrected_decimal_age_comment = "Your child was born at term. No correction for gestation has been made."
                        clinician_corrected_decimal_age_comment = "Born at term. No correction for gestation has been made."
    elif chronological_decimal_age > corrected_decimal_age or chronological_decimal_age < corrected_decimal_age:
        ## adjustment for gestational age has been made - even if >=37 weeks
        lay_corrected_decimal_age_comment = f"Because your child was born at {gestation_weeks}+{gestation_days} weeks gestation an adjustment has been made to take this into account."
        clinician_corrected_decimal_age_comment = "Correction for gestational age has been made."
        lay_chronological_decimal_age_comment = "This is your child's age without taking into account their gestation at birth."
        clinician_chronological_decimal_age_comment = "No correction has been made for gestational age."
    else:
        #some error
        lay_corrected_decimal_age_comment = "It has not been possible to calculate age this time."
        clinician_corrected_decimal_age_comment = "It has not been possible to calculate age this time."
        lay_chronological_decimal_age_comment = "It has not been possible to calculate age this time."
        clinician_chronological_decimal_age_comment = "It has not been possible to calculate age this time."

    comment = {
        'lay_corrected_comment': lay_corrected_decimal_age_comment,
        'lay_chronological_comment': lay_chronological_decimal_age_comment,
        'clinician_corrected_comment': clinician_corrected_decimal_age_comment,
        'clinician_chronological_comment': clinician_chronological_decimal_age_comment
    }
    return comment


"""
Handles CDC-specific reference data selection
"""

# standard imports
import json
from importlib import resources
from pathlib import Path

# rcpch imports
from .constants import *

"""
birth_date: date of birth
observation_date: date of observation
sex: sex (string, MALE or FEMALE)
decimal_age: chronological, decimal
corrected_age: corrected for prematurity, decimal
measurement_method: height, weight, bmi, ofc (decimal)
observation: value (float)
gestation_weeks: gestational age(weeks), integer
gestation_days: supplementary days of gestation
lms: L, M or S
reference: reference data
"""

# load the reference data
data_directory = resources.files("rcpchgrowth.data_tables")

# data_path = Path(data_directory,"fenton", "fenton.json")  # 23 weeks to 50 weeks - currently not in the code base
# with open(data_path) as json_file:
#     FENTON_DATA = json.load(json_file)
#     json_file.close()
FENTON_DATA = []

data_path = Path(data_directory, "cdc_infants.json")  # CDC interpretation of WHO 0-2y
with open(data_path) as json_file:
    CDC_INFANT_DATA = json.load(json_file)
    json_file.close()

data_path = Path(data_directory, "cdc2-20.json")  # 2 years to 20 years
with open(data_path) as json_file:
    CDC_CHILD_DATA = json.load(json_file)
    json_file.close()
# public functions


def reference_data_absent(age: float, measurement_method: str, sex: str):
    """
    Helper function.
    Returns boolean
    Tests presence of valid reference data for a given measurement request

    Reference data is not complete for all ages/sexes/measurements.
    - WHO data is used from 0-2 years
    - CDC data is used from 2-20 years
    """

    if age < 0:
        return True, "CDC data does not exist below 40 weeks."
    if age > TWENTY_YEARS:  # upper threshold of UK90 data
        return True, "CDC data does not exist above 20 years."
    if measurement_method == HEAD_CIRCUMFERENCE and age > THREE_YEARS:
        return True, "CDC data does not exist for head circumference above 3 years."
    if measurement_method == BMI and age < TWO_YEARS:
        return True, "CDC data does not exist for BMI below 2 years."
    else:
        return False, None


def cdc_reference(age: float, measurement_method, default_youngest_reference: bool = False) -> json:
    """
    The purpose of this function is to choose the correct reference for calculation.
    The CDC standard is an unusual case because it combines two different reference sources: the 1977 National Center for Health Statistics (NCHS) and the WHO2006.
    - CDC reference runs from 2.04 y to 20 y
    - WHO 2006 runs from 0  to 2 years
    - Preterm data is handled by Fenton (which is Canadian data)
    The function return the appropriate reference file as json
    """

    # These conditionals are to select the correct reference
    if age < FENTON_LOWER_THRESHOLD:
        # Below the range for which we have reference data, we can't provide a calculation.
        raise ValueError(
            "There is no reference data for ages below 22 weeks gestation."
        )
    elif age < 0:
        # Below 40 weeks, Fenton data is always used
        return FENTON_DATA
    
    elif age < 2 or (age == 2 and default_youngest_reference) or (measurement_method == HEAD_CIRCUMFERENCE and age <= 3):
        # Below 2 years, CDC interpretation of WHO is used
        return CDC_INFANT_DATA

    elif age <= CDC_UPPER_THRESHOLD:
        # CDC data is used for all children 2-20 years
        return CDC_CHILD_DATA 

    else:
        return ValueError("There is no CDC reference data above the age of 20 years.")


def cdc_lms_array_for_measurement_and_sex(
    age: float,
    measurement_method: str,
    sex: str,
    default_youngest_reference: bool = False,
) -> list:

    # selects the correct lms data array from the patchwork of references that make up UK-WHO

    # Check that the measurement requested has reference data at that age

    invalid_data, data_error = reference_data_absent(
     
        age=age, measurement_method=measurement_method, sex=sex
    )
    try:
        selected_reference = cdc_reference(
            age=age, measurement_method=measurement_method, default_youngest_reference=default_youngest_reference
        )
    except:  #  there is no reference for the age supplied
        return LookupError("There is no CDC reference for the age supplied.")


    if invalid_data:
        raise LookupError(data_error)
    else:
        return selected_reference["measurement"][measurement_method][sex]


def select_reference_data_for_cdc_chart(
    cdc_reference_name: str,
    measurement_method: str,
    sex: str,
    default_youngest_reference: bool = False,
):

    # takes a cdc reference name (see parameter constants), measurement_method and sex to return
    # reference data - NOTE for the moment Fenton is not implemented as license not yet secured and reference data therefore not committed to the repo

    if cdc_reference_name == FENTON:
        try:
            fenton_preterm_reference = cdc_lms_array_for_measurement_and_sex(
                age=-0.01, # an arbritrary age to select the preterm data
                measurement_method=measurement_method,
                sex=sex,
                default_youngest_reference=False # should never need younger reference in this calculation
            )
        except:
            fenton_preterm_reference = []
        return fenton_preterm_reference
    elif cdc_reference_name == CDC_INFANT:
        try:
            cdc_infant_reference = cdc_lms_array_for_measurement_and_sex(
                age=0.04, # an arbritrary age to select the infant data
                measurement_method=measurement_method,
                sex=sex,
                default_youngest_reference=default_youngest_reference
            )
        except:
            cdc_infant_reference = []
        return cdc_infant_reference
    else:
        try:
            cdc_children_reference = cdc_lms_array_for_measurement_and_sex(
                age=3, # an arbritrary age to select the child data
                measurement_method=measurement_method,
                sex=sex,
                default_youngest_reference=False # There is no younger reference data for CDC
            )
        except:
            cdc_children_reference = []
        return cdc_children_reference



"""
Handles WHO reference data selection
"""

# standard imports
import json
from importlib import resources
from pathlib import Path

# rcpch imports
from .constants import *

"""
birth_date: date of birth
observation_date: date of observation
sex: sex (string, MALE or FEMALE)
decimal_age: chronological, decimal
corrected_age: corrected for prematurity, decimal
measurement_method: height, weight, bmi, ofc (decimal)
observation: value (float)
gestation_weeks: gestational age(weeks), integer
gestation_days: supplementary days of gestation
lms: L, M or S
reference: reference data
"""

# load the reference data
data_directory = resources.files("rcpchgrowth.data_tables")

data_path = Path(
    data_directory, "who_infants.json")  # 2 weeks to 2 years
with open(data_path) as json_file:
    WHO_INFANTS_DATA = json.load(json_file)
    json_file.close()

data_path = Path(
    data_directory, "who_children.json")  # 2 years to 5 years
with open(data_path) as json_file:
    WHO_CHILD_DATA = json.load(json_file)
    json_file.close()

data_path = Path(
    data_directory, "who_2007_children.json")  # 5 years to 19 years
with open(data_path) as json_file:
    WHO_2007_DATA = json.load(json_file)
    json_file.close()

# public functions


def reference_data_absent(
    age: float,
    measurement_method: str,
    sex: str
):
    """
    Helper function.
    Returns boolean
    Tests presence of valid reference data for a given measurement request

    Reference data is not complete for all ages/sexes/measurements.
     - Length data is not available until 25 weeks gestation, though weight date is available from 23 weeks
     - There is only BMI reference data from 2 weeks of age to aged 20y
     - Head circumference reference data is available from 23 weeks gestation to 17y in girls and 18y in boys
     - lowest threshold is 23 weeks, upper threshold is 20y
    """

    if age < ZERO_YEARS:  # lower threshold of WHO data
        return True, "WHO data does not exist below term."

    if age > NINETEEN_YEARS:  # upper threshold of UK90 data
        return True, "WHO data does not exist above 19 years."

    if measurement_method == WEIGHT and age > TEN_YEARS:
        return True, "WHO weight data does not exist in children over 10 y of age."
    
    if measurement_method == HEAD_CIRCUMFERENCE and age > FIVE_YEARS:
        return True, "WHO head circumference data does not exist in children over 5 y of age."

    else:
        return False, ""


def who_reference(
    age: float,
    default_youngest_reference: bool = False
) -> json:
    """
    The purpose of this function is to choose the correct reference for calculation.
    The UK-WHO standard is an unusual case because it combines two different reference sources.
    - UK90 reference runs from 23 weeks to 20 y
    - WHO 2006 runs from 2 weeks to 4 years
    - UK90 then resumes from 4 years to 20 years
    The function return the appropriate reference file as json
    """

    # # These conditionals are to select the correct reference
    # if age < WHO_2006_REFERENCE_LOWER_THRESHOLD:
    #     # Below the range for which we have reference data, we can't provide a calculation.
    #     raise LookupError("There is no WHO reference data below 42 weeks gestation")


    if age <= WHO_2006_REFERENCE_UPPER_THRESHOLD:
        # Children up to and including 5 years are measured using WHO 2006 data
        if (age == 2.0 and default_youngest_reference) or age < WHO_CHILD_LOWER_THRESHOLD:
            # If default_youngest_reference is True, the younger reference is used to calculate values
            # This is specifically for the overlap between WHO 2006 lying and standing in centile curve generation
            # WHO 2006 reference is used for children below 2 years or those who are 2 years old and default_youngest_reference is True
            return WHO_INFANTS_DATA
        return WHO_CHILD_DATA
        
    elif age <= WHO_2007_REFERENCE_UPPER_THRESHOLD:
        # All children over 5 years and above are measured using WHO 2007 child data
        return WHO_2007_DATA

    else:
        raise LookupError("There are no WHO reference data above the age of 19 years.")


def who_lms_array_for_measurement_and_sex(
    age: float,
    measurement_method: str,
    sex: str,
    default_youngest_reference: bool = False
) -> list:

    # selects the correct lms data array from the patchwork of references that make up UK-WHO

    try:
        selected_reference = who_reference(
            age=age,
            default_youngest_reference=default_youngest_reference
        )
    except:  #  there is no reference for the age supplied
        raise LookupError("There is no WHO reference for the age supplied.")

    # Check that the measurement requested has reference data at that age

    invalid_data, data_error = reference_data_absent(
        age=age,
        measurement_method=measurement_method,
        sex=sex)

    if invalid_data:
        raise LookupError(data_error)
    else:
        return selected_reference["measurement"][measurement_method][sex]


def select_reference_data_for_who_chart(
    who_reference_name: str, 
    measurement_method: str, 
    sex: str,
    default_youngest_reference: bool = False):

    # takes a who_reference name (see parameter constants), measurement_method and sex to return
    # reference data

    if who_reference_name == WHO_2006_INFANT:
        try:
            who_infants_reference = who_lms_array_for_measurement_and_sex(
                age=0.04,
                measurement_method=measurement_method,
                sex=sex,
                default_youngest_reference=default_youngest_reference 
            )
        except:
            who_infants_reference = []
        return who_infants_reference
    elif who_reference_name == WHO_2006_CHILD:
        try:
            who_2006_children_reference = who_lms_array_for_measurement_and_sex(
                age=3.0,
                measurement_method=measurement_method,
                sex=sex,
                default_youngest_reference=default_youngest_reference # should never need younger reference in this calculation
            )
        except:
            who_2006_children_reference = []
        return who_2006_children_reference
    elif who_reference_name == WHO_2007_CHILD:
        try:
            who_2007_children_reference = who_lms_array_for_measurement_and_sex(
                age=6.0,
                measurement_method=measurement_method,
                sex=sex,
                default_youngest_reference=default_youngest_reference
            )
        except:
            who_2007_children_reference = []
        return who_2007_children_reference
    else:
        raise LookupError(
            f"No data found for {measurement_method} in {sex}s in {who_reference_name}")



"""
Tests for the global functions which perform lower-level intermediate calculations
"""

# standard imports
from datetime import datetime
import json
import os

# third-party imports
import pytest

# rcpch imports
from rcpchgrowth import global_functions, date_calculations

# The ACCURACY constant defines the level of accuracy required.
# Statistical functions in R (used by Tim Cole to create the input data we are testing against)
#   differ minutely from statistical function in Python. This is expected, and we use the ACCURACY
#   constant to account for this
ACCURACY = 1e-3


def load_valid_data_set():
    """
    Loads in the testing data from JSON file
    """
    with open(os.path.abspath(os.path.dirname(__file__)) + "/sds_age_validation_2021.json") as f:
        return json.load(f)


# # date_calculations
@pytest.mark.parametrize("line", load_valid_data_set())
def test_chronological_decimal_age(line):
    """
    Tests the function which calculates chronological decimal age
        from birth_date and observation_date
    """
    if line["birth_date"] is None or line["observation_date"] is None:
        return
    birth_date = datetime.strptime(line["birth_date"], "%d/%m/%Y").date()
    observation_date = datetime.strptime(
        line["observation_date"], "%d/%m/%Y").date()
    age = date_calculations.chronological_decimal_age(
        birth_date, observation_date)
    tim_age = float(line["chronological_age"])
    assert age == pytest.approx(tim_age, abs=ACCURACY)


@pytest.mark.parametrize("line", load_valid_data_set())
def test_corrected_decimal_age(line):
    """
    Tests the function which calculates **corrected** decimal age
    """
    if line["birth_date"] is None or line["observation_date"] is None:
        return
    birth_date = datetime.strptime(line["birth_date"], "%d/%m/%Y").date()
    observation_date = datetime.strptime(
        line["observation_date"], "%d/%m/%Y").date()
    age = date_calculations.corrected_decimal_age(birth_date, observation_date, int(
        line["gestation_weeks"]), int(line["gestation_days"]))
    tim_age = float(line["corrected_age"])
    assert age == pytest.approx(tim_age, abs=ACCURACY)


# SDS calculation tests
@pytest.mark.parametrize("line", load_valid_data_set())
def test_sds_for_measurement_corrected(line):
    """
    Tests the function with calculates SDS from gestationally-corrected measurements
    """
    if line["observation_value"] is None or line["corrected_sds"] is None:
        return
    sds = global_functions.sds_for_measurement("uk-who", float(line["corrected_age"]), str(
        line["measurement_method"]), float(line["observation_value"]), str(line["sex"]))
    tim_sds = float(line["corrected_sds"])
    assert sds == pytest.approx(tim_sds, abs=ACCURACY)


@pytest.mark.parametrize("line", load_valid_data_set())
def test_sds_for_measurement_chronological(line):
    """
    Tests the function with calculates SDS from uncorrected measurements
    """
    if line["observation_value"] is None or line["chronological_sds"] is None:
        return
    sds = global_functions.sds_for_measurement("uk-who", float(line["chronological_age"]), str(
        line["measurement_method"]), float(line["observation_value"]), str(line["sex"]))
    tim_sds = float(line["chronological_sds"])
    assert sds == pytest.approx(tim_sds, abs=ACCURACY)



# standard imports
from datetime import date
from typing import Literal

# rcpch imports
from .centile_bands import centile_band_for_centile
from .constants import *
from .date_calculations import (chronological_decimal_age, corrected_decimal_age,
                                chronological_calendar_age, estimated_date_delivery, corrected_gestational_age)
from .global_functions import sds_for_measurement, centile, percentage_median_bmi
from .age_advice_strings import comment_prematurity_correction
class Measurement:

    def __init__(
        self,
        birth_date: date,
        measurement_method: str,
        observation_date: date,
        observation_value: float,
        reference: str,
        sex: str,
        gestation_days: int = 0,
        gestation_weeks: int = 0,
        events_text: list = None,
        bone_age: float = None,
        bone_age_type: str = None,
        bone_age_sds: float = None,
        bone_age_centile: float = None,
        bone_age_text: str = None
    ):
        """
        The Measurement Class is the gatekeeper to all the functions in the RCPCHGrowth package, although the public
        functions can be accessed independently. The bulk of the error handling happens here so be aware that calling
        other functions independently may yield unexpected results.
        It is initialised with the following Required parameters:
        `birth_date`: (Python datetime object) The date of birth of the subject.
        `measurement_type`: (string) 'height', 'weight', 'bmi' or 'ofc' only are accepted.
        `observation_date`: (Python datetime object) The date that the observation was made.
        `observation_value`: (float) The value of the height, weight, BMI or ofc observation.
        `sex`: (string) The sex of the child, which can either be 'male' or 'female'.
        Additionally there are the following optional parameters:
        `gestation_weeks`: (integer) gestation at birth in weeks.
        `gestation_days`: (integer) supplemental days in addition to gestation_weeks at birth.
        `reference`: ENUM refering to which reference dataset to use: ['uk-who', 'turners-syndrome', 'trisomy-21', 'trisomy-21-aap', 'cdc', 'who'].
        `height_prediction`: decimal years
        `height_prediction_sds`: SDS for height prediction against reference
        `height_prediction_centile`: centile for height prediction against reference
        `height_prediction_reference`: enum ['bayley-pinneau', 'roche-wainer-thissen']
        `event_text`: list. this is a list of strings which are comments to tag a given measurement/plot with contextual information 
        `bone_age`: an estimated skeletal age calculated from xrays reported in decimal years
        `bone_age_sds`: an SDS for the bone age, based on references
        `bone_age_centile`: a centile for the bone age, based on references
        `bone_age_reference`: enum ['greulich-pyle', 'tanner-whitehouse-ii', 'tanner-whitehouse-iii', 'fels', 'bonexpert']
        """

        self.birth_date = birth_date
        self.gestation_days = gestation_days
        self.gestation_weeks = gestation_weeks
        self.measurement_method = measurement_method
        self.observation_date = observation_date
        self.observation_value = observation_value
        self.reference = reference
        self.sex = sex
        self.bone_age = bone_age
        self.bone_age_type = bone_age_type
        self.bone_age_sds = bone_age_sds
        self.bone_age_centile = bone_age_centile
        self.bone_age_text = bone_age_text
        # self.height_prediction = height_prediction
        # self.height_prediction_sds = height_prediction_sds
        # self.height_prediction_centile = height_prediction_centile
        # self.height_prediction_reference = height_prediction_reference
        self.events_text = events_text


        # the ages_object receives birth_data and measurement_dates objects
        self.ages_object = self.__calculate_ages(
            sex=self.sex,
            birth_date=self.birth_date,
            observation_date=self.observation_date,
            gestation_weeks=self.gestation_weeks,
            gestation_days=self.gestation_days)
        
        # validate the measurement method to ensure that the observation value is within the expected range - changed to SDS-based cutoffs - issue #32
        try:
            self.__validate_measurement_method(
                measurement_method=measurement_method, observation_value=observation_value, corrected_decimal_age=self.ages_object['measurement_dates']['corrected_decimal_age'], reference=reference, sex=sex)
            observation_value_error = None
        except Exception as err:
            observation_value_error = f"{err}"
        
        # the calculate_measurements_object receives the child_observation_value and measurement_calculated_values objects
        self.calculated_measurements_object = self.sds_and_centile_for_measurement_method(
            sex=self.sex,
            corrected_age=self.ages_object['measurement_dates']['corrected_decimal_age'],
            chronological_age=self.ages_object['measurement_dates']['chronological_decimal_age'],
            measurement_method=self.measurement_method,
            observation_value=self.observation_value,
            observation_value_error=observation_value_error,
            reference=self.reference
        )

        corrected_gestational_age = ""
        if (self.ages_object["measurement_dates"]["corrected_gestational_age"]["corrected_gestation_weeks"] is not None):
            corrected_gestational_age = f'{ self.ages_object["measurement_dates"]["corrected_gestational_age"]["corrected_gestation_weeks"] } + { self.ages_object["measurement_dates"]["corrected_gestational_age"]["corrected_gestation_days"]} weeks'

        self.plottable_centile_data = {
            "chronological_decimal_age_data": {
                "x": self.ages_object['measurement_dates']['chronological_decimal_age'],
                "y": self.observation_value,
                "b": self.bone_age,
                "centile": self.calculated_measurements_object["measurement_calculated_values"]["chronological_centile"],
                "sds": self.calculated_measurements_object["measurement_calculated_values"]["chronological_sds"],
                "events_text": self.events_text,
                "bone_age_label": self.bone_age_text,
                "bone_age_type": self.bone_age_type,
                "bone_age_sds": self.bone_age_sds,
                "bone_age_centile": self.bone_age_centile,
                "observation_error": self.calculated_measurements_object['child_observation_value']["observation_value_error"],
                "age_type": "chronological_age",
                "calendar_age": self.ages_object["measurement_dates"]["chronological_calendar_age"],
                "lay_comment": self.ages_object["measurement_dates"]["comments"]["lay_chronological_decimal_age_comment"],
                "clinician_comment": self.ages_object["measurement_dates"]["comments"]["clinician_chronological_decimal_age_comment"],
                "age_error": self.ages_object["measurement_dates"]["corrected_decimal_age_error"],
                "centile_band": self.calculated_measurements_object['measurement_calculated_values']["chronological_centile_band"],
                "observation_value_error": self.calculated_measurements_object["measurement_calculated_values"]["chronological_measurement_error"]

            },
            "corrected_decimal_age_data": {
                "x": self.ages_object['measurement_dates']['corrected_decimal_age'],
                "y": self.observation_value,
                "b": self.bone_age,
                "centile": self.calculated_measurements_object["measurement_calculated_values"]["corrected_centile"],
                "sds": self.calculated_measurements_object["measurement_calculated_values"]["corrected_sds"],
                "events_text": self.events_text,
                "bone_age_label": self.bone_age_text,
                "bone_age_type": self.bone_age_type,
                "bone_age_sds": self.bone_age_sds,
                "bone_age_centile": self.bone_age_centile,
                "observation_error": self.calculated_measurements_object['child_observation_value']["observation_value_error"],
                "age_type": "corrected_age",
                "corrected_gestational_age": corrected_gestational_age,
                "calendar_age": self.ages_object["measurement_dates"]["corrected_calendar_age"],
                "lay_comment": self.ages_object["measurement_dates"]["comments"]["lay_corrected_decimal_age_comment"],
                "clinician_comment": self.ages_object["measurement_dates"]["comments"]["clinician_corrected_decimal_age_comment"],
                "age_error": self.ages_object["measurement_dates"]["corrected_decimal_age_error"],
                "centile_band": self.calculated_measurements_object['measurement_calculated_values']["corrected_centile_band"],
                "observation_value_error": self.calculated_measurements_object["measurement_calculated_values"]["corrected_measurement_error"]
            }
        }

        self.plottable_sds_data = {
            "chronological_decimal_age_data": {
                "x": self.ages_object['measurement_dates']['chronological_decimal_age'],
                "y": self.calculated_measurements_object['measurement_calculated_values']["chronological_sds"],
                "b": self.bone_age,
                "centile": self.calculated_measurements_object["measurement_calculated_values"]["chronological_centile"],
                "events_text": self.events_text,
                "bone_age_label": self.bone_age_text,
                "bone_age_type": self.bone_age_type,
                "bone_age_sds": self.bone_age_sds,
                "bone_age_centile": self.bone_age_centile,
                "age_type": "chronological_age",
                "calendar_age": self.ages_object["measurement_dates"]["chronological_calendar_age"],
                "lay_comment": self.ages_object["measurement_dates"]["comments"]["lay_chronological_decimal_age_comment"],
                "clinician_comment": self.ages_object["measurement_dates"]["comments"]["clinician_chronological_decimal_age_comment"],
                "age_error": self.ages_object["measurement_dates"]["corrected_decimal_age_error"],
                "centile_band": self.calculated_measurements_object['measurement_calculated_values']["chronological_centile_band"],
                "observation_value_error": self.calculated_measurements_object["measurement_calculated_values"]["chronological_measurement_error"]
            },
            "corrected_decimal_age_data": {
                "x": self.ages_object['measurement_dates']['corrected_decimal_age'],
                "y": self.calculated_measurements_object['measurement_calculated_values']["corrected_sds"],
                "b": self.bone_age,
                "centile": self.calculated_measurements_object["measurement_calculated_values"]["corrected_centile"],
                "events_text": self.events_text,
                "bone_age_label": self.bone_age_text,
                "bone_age_type": self.bone_age_type,
                "bone_age_sds": self.bone_age_sds,
                "bone_age_centile": self.bone_age_centile,
                "age_type": "corrected_age",
                "corrected_gestational_age": corrected_gestational_age,
                "calendar_age": self.ages_object["measurement_dates"]["corrected_calendar_age"],
                "lay_comment": self.ages_object["measurement_dates"]["comments"]["lay_corrected_decimal_age_comment"],
                "clinician_comment": self.ages_object["measurement_dates"]["comments"]["clinician_corrected_decimal_age_comment"],
                "age_error": self.ages_object["measurement_dates"]["corrected_decimal_age_error"],
                "centile_band": self.calculated_measurements_object['measurement_calculated_values']["corrected_centile_band"],
                "observation_value_error": self.calculated_measurements_object["measurement_calculated_values"]["corrected_measurement_error"]
            },
        }

        # the final object is made up of these five components
        self.measurement = {
            'birth_data': self.ages_object['birth_data'],
            'measurement_dates': self.ages_object['measurement_dates'],
            'child_observation_value': self.calculated_measurements_object['child_observation_value'],
            'measurement_calculated_values': self.calculated_measurements_object['measurement_calculated_values'],
            'plottable_data': {
                "centile_data": self.plottable_centile_data,
                "sds_data": self.plottable_sds_data
            },
            # 'height_prediction_data':{
            #     "height_prediction": self.height_prediction,
            #     "height_prediction_sds": self.height_prediction_sds,
            #     "height_prediction_centile": self.height_prediction_centile,
            #     "height_prediction_reference": self.height_prediction_reference
            # },
            'bone_age': {
                "bone_age": self.bone_age,
                "bone_age_type": self.bone_age_type,
                "bone_age_sds": self.bone_age_sds,
                "bone_age_centile": self.bone_age_centile,
                "bone_age_text": self.bone_age_text,
            },
            'events_data': {
                'events_text': self.events_text
            }
        }

    """
    These are 2 public class methods
    """

    def sds_and_centile_for_measurement_method(
        self,
        sex: str,
        corrected_age: float,
        chronological_age: float,
        observation_value_error: str,
        measurement_method: str,
        observation_value: float,
        reference: str
    ):

        # returns sds for given measurement
        # bmi must be supplied precalculated

        # calculate sds based on reference, age, measurement, sex and prematurity

        if corrected_age is None or chronological_age is None:
            # there has been an age calculation error. Further calculation impossible - this may be due to a date error or because CDC reference data is not available in preterm infants

            self.return_measurement_object = self.__create_measurement_object(
                reference=reference,
                measurement_method=measurement_method,
                observation_value=observation_value,
                observation_value_error=observation_value_error,
                corrected_sds_value=None,
                corrected_centile_value=None,
                corrected_centile_band=None,
                chronological_sds_value=None,
                chronological_centile_value=None,
                chronological_centile_band=None,
                chronological_measurement_error="Dates error. Calculations impossible.",
                corrected_measurement_error="Dates error. Calculations impossible.",
                corrected_percentage_median_bmi=None,
                chronological_percentage_median_bmi=None
            )
            return self.return_measurement_object
        
            # CDC data cannot be used for preterm infants who are not yet term. We will not be able to calculate SDS scores and centiles so will return none, but signpost to the user that this is what we are doing.
        if corrected_age < 0 and reference == "cdc":
            corrected_measurement_error = "This baby is born premature. CDC data is not available for preterm infants."
            corrected_measurement_sds = None
        else:
            try:
                corrected_measurement_sds = sds_for_measurement(reference=reference, age=corrected_age, measurement_method=measurement_method,
                                                                observation_value=observation_value, sex=sex)
            except Exception as err:
                corrected_measurement_error = f"{err}"
                corrected_measurement_sds = None

        try:
            chronological_measurement_sds = sds_for_measurement(reference=reference, age=chronological_age, measurement_method=measurement_method,
                                                                observation_value=observation_value, sex=sex)
        except LookupError as err:
            chronological_measurement_error = f"{err}"
            chronological_measurement_sds = None

        if chronological_measurement_sds is None:
            chronological_measurement_centile = None
            chronological_centile_band = None
        else:
            chronological_measurement_error = None
            try:
                chronological_measurement_centile = centile(
                    z_score=chronological_measurement_sds)
            except Exception as err:
                chronological_measurement_error = "Not possible to calculate centile"
                chronological_measurement_centile = None
            try:
                if reference == CDC:
                    if measurement_method == BMI:
                        centile_format = EIGHTY_FIVE_PERCENT_CENTILES
                    else:
                        centile_format = THREE_PERCENT_CENTILES
                else:
                    centile_format = COLE_TWO_THIRDS_SDS_NINE_CENTILES
                chronological_centile_band = centile_band_for_centile(
                    sds=chronological_measurement_sds, 
                    measurement_method=measurement_method,
                    centile_format=centile_format
                    )
            except Exception as err:
                chronological_measurement_error = "Not possible to calculate centile"
                chronological_centile_band = None

        if corrected_measurement_sds is None:
            corrected_measurement_centile = None
            corrected_centile_band = None
        else:
            corrected_measurement_error = None
            try:
                corrected_measurement_centile = centile(
                    z_score=corrected_measurement_sds)
            except Exception as err:
                corrected_measurement_error = "Not possible to calculate centile"
                corrected_measurement_centile = None

            try:
                if reference == CDC:
                    if measurement_method == BMI:
                        centile_format = EIGHTY_FIVE_PERCENT_CENTILES
                    else:
                        centile_format = THREE_PERCENT_CENTILES
                else:
                    centile_format = COLE_TWO_THIRDS_SDS_NINE_CENTILES
                corrected_centile_band = centile_band_for_centile(
                    sds=corrected_measurement_sds, 
                    measurement_method=measurement_method,
                    centile_format=centile_format
                )
            except Exception as err:
                corrected_measurement_error = "Not possible to calculate centile"
                corrected_centile_band = None

        # calculate BMI centiles and percentage median BMI
        corrected_percentage_median_bmi = None
        chronological_percentage_median_bmi = None
        if measurement_method == BMI and corrected_age is not None and chronological_age is not None:
            try: 
                corrected_percentage_median_bmi = percentage_median_bmi(
                reference=reference,
                age=corrected_age,
                actual_bmi=observation_value,
                sex=sex
            )
            except Exception as err:
                print(err)
                corrected_percentage_median_bmi = None
            
            try:
                chronological_percentage_median_bmi = percentage_median_bmi(
                    reference=reference,
                    age=corrected_age,
                    actual_bmi=observation_value,
                    sex=sex
                )
            except Exception as err:
                print(err)
                chronological_percentage_median_bmi = None

        self.return_measurement_object = self.__create_measurement_object(
            reference=reference,
            measurement_method=measurement_method,
            observation_value=observation_value,
            observation_value_error=observation_value_error,
            corrected_sds_value=corrected_measurement_sds,
            corrected_centile_value=corrected_measurement_centile,
            corrected_centile_band=corrected_centile_band,
            chronological_sds_value=chronological_measurement_sds,
            chronological_centile_value=chronological_measurement_centile,
            chronological_centile_band=chronological_centile_band,
            chronological_measurement_error=chronological_measurement_error,
            corrected_measurement_error=corrected_measurement_error,
            corrected_percentage_median_bmi=corrected_percentage_median_bmi,
            chronological_percentage_median_bmi=chronological_percentage_median_bmi
        )

        return self.return_measurement_object

    """
    These are all private class methods and are only accessed by this class on initialisation
    """
    def __calculate_ages(
            self,
            sex: str,
            birth_date: date,
            observation_date: date,
            gestation_weeks: int = 0,
            gestation_days=0):

        if gestation_weeks == 0:
            # if gestation not specified, set to 40 weeks
            gestation_weeks = 40
        # calculate ages from dates and gestational ages at birth


        try:
            self.corrected_decimal_age = corrected_decimal_age(
                birth_date=birth_date,
                observation_date=observation_date,
                gestation_weeks=gestation_weeks,
                gestation_days=gestation_days)
        except Exception as err:
            self.corrected_decimal_age = None
            corrected_decimal_age_error = f"{err}"

        try:
            self.chronological_decimal_age = chronological_decimal_age(
                birth_date=birth_date,
                observation_date=observation_date)
        except Exception as err:
            self.chronological_decimal_age = None
            chronological_decimal_age_error = f"{err}"
        
        # if reference is CDC or WHO, we must treat >37 week infants as term and we also stop correcting for prematurity at 2 years of age
        if self.reference == CDC or self.reference == WHO:
            if (self.corrected_decimal_age >= 2 and gestation_weeks < 37) or (gestation_weeks >= 37 and gestation_weeks <= 42):
                self.corrected_decimal_age = self.chronological_decimal_age

        if self.corrected_decimal_age is None:
            self._age_comments = None
            self.lay_corrected_decimal_age_comment = None
            self.clinician_corrected_decimal_age_comment = None
        else:
            corrected_decimal_age_error = None
            try:
                self.age_comments = comment_prematurity_correction(
                    chronological_decimal_age=self.chronological_decimal_age,
                    corrected_decimal_age=self.corrected_decimal_age,
                    gestation_weeks=gestation_weeks,
                    gestation_days=gestation_days,
                    reference=self.reference)
            except:
                self.age_comments = None
                corrected_decimal_age_error = "Error in comment on prematurity."

            try:
                self.lay_corrected_decimal_age_comment = self.age_comments['lay_corrected_comment']
            except:
                self.lay_corrected_decimal_age_comment = None
                corrected_decimal_age_error = "Error in comment on corrected decimal age."

            try:
                self.clinician_corrected_decimal_age_comment = self.age_comments[
                    'clinician_corrected_comment']
            except:
                self.clinician_corrected_decimal_age_comment = None
                corrected_decimal_age_error = "Error in comment on corrected decimal age."

        if chronological_decimal_age is None:
            self.chronological_calendar_age = None
            self.lay_chronological_decimal_age_comment = None
            self.clinician_chronological_decimal_age_comment = None
            self.corrected_gestational_age = None
            self.estimated_date_delivery = None
            self.estimated_date_delivery_string = None
        else:
            chronological_decimal_age_error = None
            try:
                self.chronological_calendar_age = chronological_calendar_age(
                    birth_date=birth_date,
                    observation_date=observation_date)
            except:
                self.chronological_calendar_age = None
                chronological_decimal_age_error = "Chronological age calculation error."

            try:
                self.lay_chronological_decimal_age_comment = self.age_comments[
                    'lay_chronological_comment']
            except:
                self.lay_chronological_decimal_age_comment = None
                chronological_decimal_age_error = "Chronological age calculation error."

            try:
                self.clinician_chronological_decimal_age_comment = self.age_comments[
                    'clinician_chronological_comment']
            except:
                self.clinician_chronological_decimal_age_comment = None
                chronological_decimal_age_error = "Chronological age calculation error."

            try:
                self.corrected_gestational_age = corrected_gestational_age(
                    birth_date=birth_date,
                    observation_date=observation_date,
                    gestation_weeks=gestation_weeks,
                    gestation_days=gestation_days)
            except:
                self.corrected_gestational_age = None
                chronological_decimal_age_error = "Corrected gestational age calculation error."

            try:
                self.estimated_date_delivery = estimated_date_delivery(
                    birth_date, gestation_weeks, gestation_days)
            except:
                self.estimated_date_delivery = None
                self.estimated_date_delivery_string = None
                chronological_decimal_age_error = "Estimated date of delivery calculation error."

            try:
                self.corrected_calendar_age = chronological_calendar_age(
                    self.estimated_date_delivery, observation_date)
            except Exception as err:
                # The EDD is still in the future as this preterm baby is not yet term. The error returned is not useful as the function is expecting a birth date in the past but is being passed an EDD in the future.
                # It is not really an error, but a limitation of the function. The calendar age really is the same as the corrected gestational age here.
                self.corrected_calendar_age = None
                chronological_decimal_age_error = None

            try:
                self.estimated_date_delivery_string = self.estimated_date_delivery.strftime(
                    '%a %d %B, %Y')
            except:
                self.estimated_date_delivery_string = None
                chronological_decimal_age_error = "Estimated date of delivery calculation error."

        birth_data = {
            "birth_date": birth_date,
            "gestation_weeks": gestation_weeks,
            "gestation_days": gestation_days,
            "estimated_date_delivery": self.estimated_date_delivery,
            "estimated_date_delivery_string": self.estimated_date_delivery_string,
            "sex": sex
        }

        measurement_dates = {
            "observation_date": observation_date,
            "chronological_decimal_age": self.chronological_decimal_age,
            "corrected_decimal_age": self.corrected_decimal_age,
            "chronological_calendar_age": self.chronological_calendar_age,
            "corrected_calendar_age": self.corrected_calendar_age,
            "corrected_gestational_age": {
                "corrected_gestation_weeks": self.corrected_gestational_age["corrected_gestation_weeks"],
                "corrected_gestation_days": self.corrected_gestational_age["corrected_gestation_days"],
            },
            "comments": {
                "clinician_corrected_decimal_age_comment": self.clinician_corrected_decimal_age_comment,
                "lay_corrected_decimal_age_comment": self.lay_corrected_decimal_age_comment,
                "clinician_chronological_decimal_age_comment": self.clinician_chronological_decimal_age_comment,
                "lay_chronological_decimal_age_comment": self.lay_chronological_decimal_age_comment
            },
            "corrected_decimal_age_error": corrected_decimal_age_error,
            "chronological_decimal_age_error": chronological_decimal_age_error
        }

        child_age_calculations = {
            "birth_data": birth_data,
            "measurement_dates": measurement_dates
        }
        return child_age_calculations

    def __create_measurement_object(
        self,
        reference: str,
        measurement_method: str,
        observation_value: float,
        observation_value_error: str,
        corrected_sds_value: float,
        corrected_centile_value: float,
        corrected_centile_band: str,
        chronological_sds_value: float,
        chronological_centile_value: float,
        chronological_centile_band: str,
        chronological_measurement_error: str,
        corrected_measurement_error: str,
        corrected_percentage_median_bmi: str,
        chronological_percentage_median_bmi: str
    ):
        """
        private class method
        This is the end step, having calculated dates, SDS/Centiles and selected appropriate clinical advice,
        to then create a bespoke json Measurement object with values relevant only to the measurement_method requested
        @params: measurement_method: string accepting only 'height', 'weight', 'bmi', 'ofc' lowercase only
        """

        # Measurement object is made up of 4 JSON elements: "birth_data", "measurement_dates",
        #  "child_observation_value" and "measurement_calculated_values"
        # All Measurement objects return the "birth_data" and "measurement_dates" elements
        # Only those calculations relevant to the measurement_method requested populate the final JSON
        # object.

        measurement_calculated_values = {
            "corrected_sds": corrected_sds_value,
            "corrected_centile": corrected_centile_value,
            "corrected_centile_band": corrected_centile_band,
            "chronological_sds": chronological_sds_value,
            "chronological_centile": chronological_centile_value,
            "chronological_centile_band": chronological_centile_band,
            "corrected_measurement_error": corrected_measurement_error,
            "chronological_measurement_error": chronological_measurement_error,
            "corrected_percentage_median_bmi": corrected_percentage_median_bmi,
            "chronological_percentage_median_bmi":chronological_percentage_median_bmi
        }

        child_observation_value = {
            "measurement_method": measurement_method,
            "observation_value": observation_value,
            "observation_value_error": observation_value_error
        }

        return {
            "child_observation_value": child_observation_value,
            "measurement_calculated_values": measurement_calculated_values,
        }

    def __validate_measurement_method(
            self,
            measurement_method: str,
            observation_value: float,
            corrected_decimal_age: float,
            sex: Literal["male", "female"],
            reference: Literal['uk-who', 'turners-syndrome', 'trisomy-21', 'trisomy-21-aap', 'cdc', 'who'] = 'uk-who'):

        # Private method which accepts a measurement_method (height, weight, bmi or ofc), reference and age as well as observation value
        # and returns True if valid

        is_valid = False

        observation_value_z_score = None
        if observation_value is not None:
            observation_value_z_score = sds_for_measurement(
                reference=reference, age=corrected_decimal_age, measurement_method=measurement_method, observation_value=observation_value, sex=sex)

        if measurement_method == 'bmi':
            if observation_value is None:
                raise ValueError(
                    'Missing observation_value for Body Mass Index. Please pass a Body Mass Index in kilograms per metre squared (kg/m²)')
            elif observation_value_z_score < MINIMUM_BMI_ERROR_SDS:
                raise ValueError(
                    f'The Body Mass Index measurement of {observation_value} kg/m² is below -15 SD and considered to be an error.')
            elif observation_value_z_score > MAXIMUM_BMI_ERROR_SDS:
                raise ValueError(
                    f'The Body Mass Index measurement of {observation_value} kg/m² is above +15 SD and considered to be an error.')
            else:
                is_valid = True

        elif measurement_method == 'height':
            if observation_value is None:
                raise ValueError(
                    'Missing observation_value for height/length. Please pass a height/length in cm.')
            elif observation_value < 2:
                # most likely metres passed instead of cm.
                raise ValueError(
                    'Height/length must be passed in cm, not metres')
            elif observation_value_z_score < MINIMUM_HEIGHT_WEIGHT_OFC_ERROR_SDS:
                raise ValueError(
                    f'The height/length of {observation_value} cm is below -8 SD and considered to be an error.')
            elif observation_value_z_score > MAXIMUM_HEIGHT_WEIGHT_OFC_ERROR_SDS:
                raise ValueError(
                    f'The height/length of {observation_value} cm is above +8 SD and considered to be an error.')
            else:
                is_valid = True

        elif measurement_method == 'weight':
            if observation_value is None:
                raise ValueError(
                    'Missing observation_value for weight. Please pass a weight in kilograms.')
            elif observation_value_z_score < MINIMUM_HEIGHT_WEIGHT_OFC_ERROR_SDS:
                raise ValueError(
                    f'The weight of {observation_value} kg is below -8 SD and considered to be an error.')
            elif observation_value_z_score > MAXIMUM_HEIGHT_WEIGHT_OFC_ERROR_SDS:
                # it is likely the weight is passed in grams, not kg.
                raise ValueError(
                    f'The weight of {observation_value} kg is above +8 SD and considered to be an error. Note that the weight should be supplied in kilograms.')
            else:
                is_valid = True

        elif measurement_method == 'ofc':
            if observation_value is None:
                raise ValueError(
                    'Missing observation_value for head circumference. Please pass a head circumference in centimetres.')
            elif observation_value_z_score < MINIMUM_HEIGHT_WEIGHT_OFC_ERROR_SDS:
                raise ValueError(
                    f'The head circumference of {observation_value} cm is below -8 SD and considered to be an error.')
            elif observation_value_z_score > MAXIMUM_HEIGHT_WEIGHT_OFC_ERROR_SDS:
                raise ValueError(
                    f'The head circumference of {observation_value} cm is above +8 SD and considered to be an error.')
            else:
                is_valid = True

        return is_valid


from .age_advice_strings import comment_prematurity_correction
from .bmi_functions import bmi_from_height_weight, weight_for_bmi_height
from .cdc import select_reference_data_for_cdc_chart
from .centile_bands import centile_band_for_centile
from .chart_functions import create_chart, generate_custom_sds_line
from .constants import *
from .date_calculations import chronological_decimal_age, corrected_decimal_age, chronological_calendar_age, estimated_date_delivery, corrected_gestational_age
from .dynamic_growth import create_thrive_line, return_correlation, create_thrive_lines
from .global_functions import centile, sds_for_measurement, measurement_from_sds, percentage_median_bmi, measurement_for_z, cubic_interpolation, linear_interpolation
from .fictional_child import generate_fictional_child_data
from .measurement import Measurement
from .mid_parental_height import mid_parental_height, mid_parental_height_z, expected_height_z_from_mid_parental_height_z, lower_and_upper_limits_of_expected_height_z
from .trisomy_21 import select_reference_data_for_trisomy_21
from .trisomy_21_aap import select_reference_data_for_trisomy_21_aap
from .turner import select_reference_data_for_turner
from .uk_who import select_reference_data_for_uk_who_chart



"""
Reference constants
"""

# Trisomy 21 constants
TRISOMY_21 = "trisomy-21"
TRISOMY_21_AAP = "trisomy-21-aap"
TRISOMY_21_AAP_INFANT = "trisomy_21_aap_infant"
TRISOMY_21_AAP_CHILD = "trisomy_21_aap_child"
TRISOMY_21_AAP_REFERENCES = [TRISOMY_21_AAP_INFANT, TRISOMY_21_AAP_CHILD]

# Turner constants
TURNERS = "turners-syndrome"

# UK-WHO constants
UK_WHO = "uk-who"
UK90_PRETERM = "uk90_preterm"
UK_WHO_INFANT = "uk_who_infant"
UK_WHO_CHILD = "uk_who_child"
UK90_CHILD = "uk90_child"
UK_WHO_REFERENCES = [UK90_PRETERM, UK_WHO_INFANT, UK_WHO_CHILD, UK90_CHILD]

# CDC constants
CDC = "cdc" # CDC is the overarching reference
CDC_INFANT = "cdc_infant" # CDC infant is the reference name for children 0-2 years
CDC_CHILD = "cdc_child" # CDC child is the reference name for children 2-20 years
FENTON = "fenton" # Fenton is the reference name for preterm infants
CDC_REFERENCES = [FENTON, CDC_INFANT, CDC_CHILD] # CDC references

# WHO constants
WHO = "who" # WHO is the overarching reference
WHO_2006_INFANT = "who_2006_infant" # WHO 2006 infant is the reference name for children 0-2 years
WHO_2006_CHILD = "who_2006_child" # WHO 2006 child is the reference name for children 2-5 years
WHO_2007_CHILD = "who_2007_child" # WHO 2007 child is the reference name for children 5-19 years
WHO_REFERENCES = [WHO_2006_INFANT, WHO_2006_CHILD, WHO_2007_CHILD] # WHO references

WHO_2006_REFERENCE_LOWER_THRESHOLD = ((42 * 7) - (40 * 7)) / 365.25  # 42 weeks as decimal age  # 2 weeks as decimal age
WHO_2006_REFERENCE_UPPER_THRESHOLD = 5.0  # 5 years as decimal age
WHO_2007_REFERENCE_LOWER_THRESHOLD = 5.0  # 5 years as decimal age
WHO_2007_REFERENCE_UPPER_THRESHOLD = 19.0  # 19 years as decimal age

# 23 weeks is the lowest decimal age available on the UK90 charts
UK90_REFERENCE_LOWER_THRESHOLD = (
    (23 * 7) - (40 * 7)
) / 365.25  # 23 weeks as decimal age

# The WHO references change from measuring infants in the lying position to measuring children in the standing position at 2.0 years.
WHO_CHILD_LOWER_THRESHOLD = 2.0  # 2 years as decimal age
# The UK-WHO standard is complicated because it switches from the WHO references to UK90 references
#  at the age of 4.0 years. This is because it was felt the reference data from breast fed infants
#  from the WHO cohorts were more accurate than the UK90 cohorts for this age group.
#  The Term reference averaged all L, M and S from 37-42 weeks. This is now deprecated and therefore UK90 data is used
# for all measurements across this age range
# Caution is advised when interpreting serial measurements over this time periods - babies are often measured inaccurately and
# up to 10% weight loss is expected in the first 2 weeks of life, and birthweight is often not regained until
# 3 weeks of life

WHO_CHILDREN_UPPER_THRESHOLD = 4.0
UK_WHO_INFANT_LOWER_THRESHOLD = (
    (42 * 7) - (40 * 7)
) / 365.25  # 42 weeks as decimal age
UK90_UPPER_THRESHOLD = 20

WHO_NEWBORN_LOWER_THRESHOLD = 0.0
CDC_LOWER_THRESHOLD = 2.0
CDC_UPPER_THRESHOLD = 20.0

FENTON_LOWER_THRESHOLD = ((22 * 7) - (40 * 7)) / 365.25  # 22 weeks as decimal age
FENTON_UPPER_THRESHOLD = ((50 * 7) - (40 * 7)) / 365.25  # 50 weeks as decimal age


# Generic constants
MALE = "male"
FEMALE = "female"
HEIGHT = "height"
WEIGHT = "weight"
HEAD_CIRCUMFERENCE = "ofc"
BMI = "bmi"
REFERENCES = [UK_WHO, TRISOMY_21, TRISOMY_21_AAP, TURNERS, CDC, WHO]
SEXES = [MALE, FEMALE]
MEASUREMENT_METHODS = [HEIGHT, WEIGHT, HEAD_CIRCUMFERENCE, BMI]

# Nomenclature
# Centile formats describe the collection - the two most common used are the cole-nine-centiles format used in UK-WHO
# Each centile in the cole format is 2/3 of a standard deviation apart. They map to 0.04/2/9/25/50/75/91/98/99.6

# and the three-percent-centiles and five-percent-centiles are used by the CDC. They use 5/10/25/50/75/90/95 but also replace the 5th centile with a 3rd/97th centile..
# The eight-five-percent-centiles format is also used by the CDC and adds an extra 85th centile as well as extended centiles for BMI  [5.0, 10.0, 25.0, 50.0, 75.0, 85, 90.0, 98.0, 99.0, 99.9, 99.99 centiles] for use in BMI calculations - this is not used in the UK-WHO references. 
# They were introduced in December 2022

THREE_PERCENT_CENTILES = "three-percent-centiles"
FIVE_PERCENT_CENTILES = "five-percent-centiles"
EIGHTY_FIVE_PERCENT_CENTILES = "eighty-five-percent-centiles"
COLE_TWO_THIRDS_SDS_NINE_CENTILES = "cole-nine-centiles"
CENTILE_FORMATS = [THREE_PERCENT_CENTILES, COLE_TWO_THIRDS_SDS_NINE_CENTILES]

THREE_PERCENT_CENTILE_COLLECTION = [3.0, 5.0, 10.0, 25.0, 50.0, 75.0, 90.0, 95.0, 97.0]
EIGHTY_FIVE_PERCENT_CENTILE_COLLECTION = [5.0, 10.0, 25.0, 50.0, 75.0, 85.0, 90.0, 95, 98.0, 99.0, 99.9, 99.99] # use for CDC Extended BMI centiles 2022
FIVE_PERCENT_CENTILE_COLLECTION = [5.0, 10.0, 25.0, 50.0, 75.0, 90.0, 95.0]
COLE_TWO_THIRDS_SDS_NINE_CENTILE_COLLECTION = [
    0.4,
    2.0,
    9.0,
    25.0,
    50.0,
    75.0,
    91.0,
    98.0,
    99.6,
]



TERM_PREGNANCY_LENGTH_DAYS = 40 * 7
TERM_LOWER_THRESHOLD_LENGTH_DAYS = 37 * 7
EXTREME_PREMATURITY_THRESHOLD_LENGTH_DAYS = 32 * 7
TWENTY_TWO_WEEKS_GESTATION = ((22 * 7) - (40 * 7)) / 365.25  # 22 weeks as decimal age
TWENTY_FIVE_WEEKS_GESTATION = ((25 * 7) - (40 * 7)) / 365.25  # 25 weeks as decimal age
THIRTY_SEVEN_WEEKS_GESTATION = ((37 * 7) - (40 * 7)) / 365.25  # 37 weeks as decimal age
FORTY_TWO_WEEKS_GESTATION = ((42 * 7) - (40 * 7)) / 365.25  # 42 weeks as decimal age
TWENTY_THREE_WEEKS_GESTATION = ((23 * 7) - (40 * 7)) / 365.25  # 23 weeks as decimal age
FIFTY_WEEKS_GESTATION = ((50 * 7) - (40 * 7)) / 365.25  # 50 weeks as decimal age
ZERO_YEARS = 0.0
TWO_YEARS = 2.0
THREE_YEARS = 3.0
FIVE_YEARS = 5.0
TEN_YEARS = 10.0
SEVENTEEN_YEARS = 17.0
EIGHTEEN_YEARS = 18.0
NINETEEN_YEARS = 19.0
TWENTY_YEARS = 20.0
TERM_WEEKS = 40
LOWER_THRESHOLD_PRETERM_REFERENCE_WEEKS = 23



# standard imports
import json
from typing import Literal
import os
import math

# from scipy.interpolate import UnivariateSpline
# import matplotlib.pyplot as plt
# TODO #21 - do the above lines need to be commented out?

# rcpchgrowth local imports
from rcpchgrowth.constants.reference_constants import FEMALE, MALE, UK_WHO, WEIGHT
from rcpchgrowth.global_functions import measurement_from_sds, sds_for_centile, z_score, linear_interpolation

"""
These functions are experimental
Height, weight, BMI or OFC in terms of SDS / Centile are snapshots in time and tell
us actually very little about growth, which is a dynamic measure. In order to make
predictions, we need to look at change in parameter measured over time (velocity)
which requires 2 measurements over a known time interval, or change in velocity (acceleration/deceleration)
which requires three measurements.

From these measurements predictions can be made about speed of growth, or rate of slowing (catch down)
or acceleration (catch up). The normative data against which to compare the index child are
thrive lines, generated here.

"""


def velocity(parameter: str, measurements_array):
    """
    This is an experimental function and not to be used clinically because velocity is not constant
    and is age dependent.
    Velocity needs at least 2 measurements from 2 consecutive time points.
    This takes an array of Measurement objects of the same child, removes the last 2 values of the same
    measurement and calculates the velocity in units/y.
    """
    parameter_list = []
    if len(measurements_array) < 2:
        return 'Not enough data'
    else:
        for measurement in measurements_array:
            if measurement:
                if parameter == measurement['child_observation_value']['measurement_method']:
                    parameter_list.append(measurement)
        if len(parameter_list) < 2:
            return f"There are not enough {parameter} values to calculate a velocity."
        else:
            last = parameter_list[-1]
            penultimate = parameter_list[-2]
            time_elapsed = last['measurement_dates']['chronological_decimal_age'] - \
                penultimate['measurement_dates']['chronological_decimal_age']
            parameter_difference = 0.0
            parameter_difference = last['child_observation_value']['observation_value'] - \
                penultimate['child_observation_value']['observation_value']
            return parameter_difference / time_elapsed


def acceleration(parameter: str, measurements_array):
    """
    This is an experimental function and not to be used clinically because acceleration is not constant
    and is age dependent.
    Accelaration needs at least 3 measurements over 3 consecutive time points in order to compare the velocity
    change between the first pair and the last pair.
    This takes an array of Measurement objects of the same child, removes the last 3 values of the same
    measurement and calculates the acceleration.
    The parameter in question is one of 'height', 'weight', 'bmi', 'ofc'
    """
    parameter_list = []
    if len(measurements_array) < 3:
        return 'Not enough data'
    else:
        for measurement in measurements_array:
            if measurement:
                if parameter == measurement['child_observation_value']['measurement_method']:
                    parameter_list.append(measurement)
        if len(parameter_list) < 3:
            return f"There are not enough {parameter} values to calculate acceleration."
        else:
            last = parameter_list[-1]
            penultimate = parameter_list[-2]
            antepentultimate = parameter_list[-3]
            first_parameter_pair_time_elapsed = penultimate['measurement_dates']['chronological_decimal_age'] - \
                antepentultimate['measurement_dates']['chronological_decimal_age']
            last_parameter_pair_time_elapsed = last['measurement_dates']['chronological_decimal_age'] - \
                penultimate['measurement_dates']['chronological_decimal_age']
            first_parameter_pair_difference = 0.0
            last_parameter_pair_difference = 0.0

            last_parameter_pair_difference = last['child_observation_value']['observation_value'] - \
                penultimate['child_observation_value']['observation_value']
            first_parameter_pair_difference = penultimate['child_observation_value']['observation_value'] - \
                antepentultimate['child_observation_value']['observation_value']

            latest_velocity = last_parameter_pair_difference / last_parameter_pair_time_elapsed
            penultimate_velocity = first_parameter_pair_difference / \
                first_parameter_pair_time_elapsed
            acceleration = (latest_velocity - penultimate_velocity) / \
                last_parameter_pair_time_elapsed
            return acceleration

def create_pairs(measurements_array: list = []):
    # test data
    """
    measurements = [{'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 25 Apr 1759 00:00:00 GMT', 'chronological_decimal_age': 0.038329911019849415, 'corrected_decimal_age': 0.038329911019849415, 'chronological_calendar_age': '2 weeks', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 4.111224921050807, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.0020115566532506, 'weight_centile': 84.18309943393204, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 09 May 1759 00:00:00 GMT', 'chronological_decimal_age': 0.07665982203969883, 'corrected_decimal_age': 0.07665982203969883, 'chronological_calendar_age': '4 weeks', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 4.699038339425533, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.002339620693291, 'weight_centile': 84.19102035307033, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 23 May 1759 00:00:00 GMT', 'chronological_decimal_age': 0.11498973305954825, 'corrected_decimal_age': 0.11498973305954825, 'chronological_calendar_age': '1 month, 1 week and 5 days', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 5.232213641808542, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.0045936184500082, 'weight_centile': 84.24537143099158, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 06 Jun 1759 00:00:00 GMT', 'chronological_decimal_age': 0.15331964407939766, 'corrected_decimal_age': 0.15331964407939766, 'chronological_calendar_age': '1 month, 3 weeks and 5 days', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 5.692927141647227, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.004963324529577, 'weight_centile': 84.25427448883711, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 20 Jun 1759 00:00:00 GMT', 'chronological_decimal_age': 0.19164955509924708, 'corrected_decimal_age': 0.19164955509924708, 'chronological_calendar_age': '2 months, 1 week and 2 days', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 6.099306464415924, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.007109988748447, 'weight_centile': 84.30590392040097, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 04 Jul 1759 00:00:00 GMT', 'chronological_decimal_age': 0.2299794661190965, 'corrected_decimal_age': 0.2299794661190965, 'chronological_calendar_age': '2 months, 3 weeks and 2 days', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 6.461379695081464, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.007475030461087, 'weight_centile': 84.31467244800477, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 18 Jul 1759 00:00:00 GMT', 'chronological_decimal_age': 0.2683093771389459, 'corrected_decimal_age': 0.2683093771389459, 'chronological_calendar_age': '3 months and 1 week', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 6.789203770743711, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.009650708766674, 'weight_centile': 84.36686671210703, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 01 Aug 1759 00:00:00 GMT', 'chronological_decimal_age': 0.3066392881587953, 'corrected_decimal_age': 0.3066392881587953, 'chronological_calendar_age': '3 months and 3 weeks', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 7.088501064489439, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.0108086749449121, 'weight_centile': 84.39459948388198, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 15 Aug 1759 00:00:00 GMT', 'chronological_decimal_age': 0.34496919917864477, 'corrected_decimal_age': 0.34496919917864477, 'chronological_calendar_age': '4 months', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 7.363720070832382, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.0126387519444007, 'weight_centile': 84.4383628580822, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 29 Aug 1759 00:00:00 GMT', 'chronological_decimal_age': 0.38329911019849416, 'corrected_decimal_age': 0.38329911019849416, 'chronological_calendar_age': '4 months, 2 weeks and 4 days', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 7.614522747033926, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.0139857964039922, 'weight_centile': 84.47052350872625, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}]
    """
    parameter_list = []
    if len(measurements_array) < 2:
        return 'Not enough data'
    else:
        for measurement in measurements_array:
            if measurement:
                if measurement['child_measurement_value']['weight'] is not None:
                    parameter_list.append(measurement)

        if len(parameter_list) < 2:
            return f"There are not enough weight values to calculate a velocity."
        else:
            last = parameter_list[-1]
            penultimate = parameter_list[-2]
            penultimate_weight_sds_value = penultimate['measurement_calculated_values']['weight_sds']
            penultimate_decimal_age = penultimate['measurement_dates']['chronological_decimal_age']
            last_weight_sds_value = last['measurement_calculated_values']['weight_sds']
            last_decimal_age = last['measurement_dates']['chronological_decimal_age']
            return {
                "penultimate_weight_sds_value": penultimate_weight_sds_value,
                "penultimate_decimal_age": penultimate_decimal_age,
                "last_weight_sds_value": last_weight_sds_value,
                "last_decimal_age": last_decimal_age
            }

# centile creation

def nine_centiles(sex: str):
    t=[centile * 0.5 for centile in range(24)]
    requested_sd_scores=[-2.67, -2.0, -1.33, -0.67, 0, 0.67, 1.33, 2.0, 2.67]
    final_array=[]

    for index, requested_sd_score in enumerate(requested_sd_scores):
        return_ages_array=[]
        return_m_array=[]
        for index in range(len(t)-1):
            m=measurement_from_sds(
                reference=UK_WHO,
                requested_sds=requested_sd_score,
                measurement_method=WEIGHT,
                sex=sex,
                age=t[index]/12
            )
            return_m_array.append(m)
            return_ages_array.append(t[index]/12)
        final_array.append({
            "ages": return_ages_array,
            "observation_values": return_m_array
        })
    return final_array

# conditional weight gain functions

def conditional_weight_gain(z1, r, Z):
    """
    Weight velocity of the individual child cannot be predicted without comparison against reference data velocity since weight velocity is age dependent. This uses a correlation matrix to look up values against which to compare the rate at which the child is gaining or losing weight.
    The formula for conditional weight gain is: (z2 – r x z1) / √1-r^2
    Conditional reference charts to assess weight gain in British infants, T J Cole, Archives of Disease in Childhood 1995; 73: 8-16f
    """
    return z1 * r + Z * math.sqrt(1 - r**2)

# create a single thrive line

def create_thrive_line(t: list, z1: float, sex: str, target_centile: float = 5.0):
    # creates a single thrive line
    # accepts a list of ages against which the measurements are plotted
    # z1 refers to the starting SDS
    # target_centile refers to the velocity centile requested (defaults to 5th centile)

    zv=sds_for_centile(target_centile)
    observation_value=None

    observation_value=measurement_from_sds(
            reference=UK_WHO,
            requested_sds=z1,
            sex=sex,
            age=t[0]/12,
            measurement_method=WEIGHT
        )
    return_observation_values=[observation_value]
    cycle_sds=[z1]
    return_ages=[t[0]]
    z2=0.0

    for index in range(len(t)):
        observation_value=None
        # loop through the list of ages which are ordered and evenly spaced a month apart
        z_current = cycle_sds[index]

        if index < len(t)-1:
            t1, t2=t[index], t[index+1]
            # use the current and then next age in the list to look up the correlation r
            r=return_correlation(t1=t1, t2=t2, time_interval="months")
            # calculate the expected z, based on requested velocity centile (zv) for the next age in the list using r
            z2=conditional_weight_gain(z_current, r, zv)
        else:
            # this is the last item in the list
            t2 = t1
            z2=z_current

        # convert z2 to a measurement and add to the list against t2 in years
        observation_value = measurement_from_sds(
            reference=UK_WHO,
            requested_sds=z2,
            sex=sex,
            age=t2/12,
            measurement_method=WEIGHT
        )

        cycle_sds.append(z2)
        return_ages.append(t2/12)
        return_observation_values.append(observation_value)

    return {
        "zs": cycle_sds,
        "ages": return_ages,
        "observation_values": return_observation_values
    }

def create_thrive_lines(target_centile: float, sex: str):
    # Creates thrive lines for weights in the under 1s
    # Each thrive line requires a starting SDS and list of ages against
    # which the lines are plotted.
    # The centile_target refers to the velocity centile cut off at which
    # the line is drawn.

    time_blocks = [i for i in range(12)]

    centile_lines=nine_centiles(sex=sex)
    for index, centile_line in enumerate(centile_lines):
        # these are the 9 centile lines
        plt.plot(centile_line["ages"], centile_line["observation_values"], linestyle='--' if index % 2==0 else '-' , color="grey", linewidth=0.5)

    z=-25
    while z <= 25:
        thrive_line = create_thrive_line(
            t=time_blocks,
            z1=z,
            sex=sex,
            target_centile=target_centile)
        z+=0.67

        # loop through the zs and remove those outside the 99.6th and 0.4th centiles. Interpolate to generate those values on the lines
        ages, measurements = [], []
        counter=0
        while counter < len(thrive_line['zs'])-1:
            if (thrive_line['zs'][counter] >= 2.667 and thrive_line['zs'][counter+1] < 2.667):
                try:
                    interpolated_age=float(linear_interpolation(2.667, thrive_line['zs'][counter+1], thrive_line['zs'][counter], thrive_line['ages'][counter+1], thrive_line['ages'][counter]))
                    interpolated_measurement = measurement_from_sds(
                        reference=UK_WHO,
                        requested_sds=2.667,
                        measurement_method=WEIGHT,
                        sex=sex,
                        age=interpolated_age
                    )
                    measurements.append(interpolated_measurement)
                    ages.append(interpolated_age)
                except Exception:
                    print("exception")
                    pass
            elif (thrive_line['zs'][counter] >= -2.667 and thrive_line['zs'][counter+1] < -2.667):
                try:
                    interpolated_age=float(linear_interpolation(-2.667, thrive_line['zs'][counter+1], thrive_line['zs'][counter], thrive_line['ages'][counter+1], thrive_line['ages'][counter]))
                    interpolated_measurement = measurement_from_sds(
                        reference=UK_WHO,
                        requested_sds=-2.667,
                        measurement_method=WEIGHT,
                        sex=sex,
                        age=interpolated_age
                    )
                    measurements.append(interpolated_measurement)
                    ages.append(interpolated_age)
                except Exception:
                    print("another exception")
                    pass
            elif thrive_line['zs'][counter] > -2.667 and thrive_line['zs'][counter] < 2.667:
                measurements.append(thrive_line['observation_values'][counter])
                ages.append(thrive_line['ages'][counter])
            counter += 1

    #     x, y = ages, measurements
    #     # smoothed_measurements = UnivariateSpline(x,y) smoothed_measurements(x)
    #     if len(x) > 0 and len(y) > 0:
    #         plt.plot(x, y, color="green", linestyle='--', dashes=(15,2.5), linewidth=1.0)

    # plt.title('Thrive Lines', fontsize=16)
    # plt.show()

def return_correlation(t1, t2, time_interval: Literal["weeks", "months"]):
    # import the reference - the index of each row and column
    # relates to the number of weeks or months

    cwd = os.path.dirname(__file__)  # current location
    if time_interval == "weeks":
        if t1 > 53 or t2 > 53:
            return Exception("Data only available below 53 weeks of age")
        file_path = 'data_tables/uk_who_weight_correlation_matrices/weight_correlation_by_week.json'
    else:
        if t1 > 12 or t2 > 12:
            return Exception("Data only available below 12 months of age")
        file_path = 'data_tables/uk_who_weight_correlation_matrices/weight_correlation_by_month.json'

    final_file_path = os.path.join(cwd, file_path)
    with open(final_file_path, mode="r") as json_file:
        data=json.load(json_file)
        json_file.close()

    lowerIndexT1 = math.floor(t1)
    upperIndexT1 = lowerIndexT1+1
    lowerIndexT2 = math.floor(t2)
    upperIndexT2 = lowerIndexT2+1

    lowerT1lowerT2 = data[lowerIndexT1][str(lowerIndexT2)]
    lowerT1upperT2 = data[lowerIndexT1][str(upperIndexT2)]
    upperT1lowerT2 = data[upperIndexT1][str(lowerIndexT2)]
    upperT1upperT2 = data[upperIndexT1][str(upperIndexT2)]

    correlation = bilinear_interpolation(
            x=t1,
            y=t2,
            points=[
                (lowerIndexT1, lowerIndexT2, lowerT1lowerT2),
                (lowerIndexT1, upperIndexT2, lowerT1upperT2),
                (upperIndexT1, lowerIndexT2, upperT1lowerT2),
                (upperIndexT1, upperIndexT2, upperT1upperT2),
            ]
        )
    return correlation


def bilinear_interpolation(x, y, points):
    '''Interpolate (x,y) from values associated with four points.

    The four points are a list of four triplets:  (x, y, value).
    The four points can be in any order.  They should form a rectangle.

        >>> bilinear_interpolation(12, 5.5,
        ...                        [(10, 4, 100),
        ...                         (20, 4, 200),
        ...                         (10, 6, 150),
        ...                         (20, 6, 300)])
        165.0

    '''
    # See formula at:  http://en.wikipedia.org/wiki/Bilinear_interpolation
    # Thanks to Raymond Hettinger on Stack overflow for this solution: https://stackoverflow.com/questions/8661537/how-to-perform-bilinear-interpolation-in-python

    points = sorted(points)               # order points by x, then by y
    (x1, y1, q11), (_x1, y2, q12), (x2, _y1, q21), (_x2, _y2, q22) = points

    if x1 != _x1 or x2 != _x2 or y1 != _y1 or y2 != _y2:
        raise ValueError('points do not form a rectangle')
    if not x1 <= x <= x2 or not y1 <= y <= y2:
        raise ValueError('(x, y) not within the rectangle')

    return (q11 * (x2 - x) * (y2 - y) +
            q21 * (x - x1) * (y2 - y) +
            q12 * (x2 - x) * (y - y1) +
            q22 * (x - x1) * (y - y1)
           ) / ((x2 - x1) * (y2 - y1) + 0.0)



"""
These constants are intended to be used throughout both the rcpchgrowth module
and the API for validation of input values and rejection of very high or low values

The constants used previously drew on the Guiness Book of Records and other sources to set hard limits.
These were replaced with SDS sds_advisory and sds_error values in the rcpchgrowth-python repository as per issue #32
"""

# HEIGHT CONSTANTS
# The smallest baby was report by the BBC on 9/8/2021 (https://www.bbc.co.uk/news/world-asia-58141756.amp)
# to have a length of just 24cm and was born at 25 weeks gestation, discharged at 13 mths of age (chronological)
# from Singapore's National University Hospital
# The tallest person for whom there is irrefutable evidence is
# Robert Pershing Wadlow who was 2.72m (8 ft 11.1 in) tall at his death in 1940
# https://en.wikipedia.org/wiki/Robert_Wadlow

# WEIGHT CONSTANTS
# The smallest baby was report by the BBC on 9/8/2021 (https://www.bbc.co.uk/news/world-asia-58141756.amp)
# to weigh just 212g and was born at 25 weeks gestation, discharged at 13 mths of age (chronological)
# from Singapore's National University Hospital
# The heaviest man according to google is Jon Brower Minnoch (US)
# who had suffered from obesity since childhood. In September 1976,
# he measured 185 cm (6 ft 1 in) tall and weighed 442 kg (974 lb; 69 st 9 lb)
# https://en.wikipedia.org/wiki/Jon_Brower_Minnoch

# OFC CONSTANTS
# We couldn't find a good answer for the 'largest human head' that we could use
# Instead have presumed that 100 is a good maximum, given that pathological states
# causing very large head size (eg hydrocephalus) render the OFC meaningless for
# growth measurement and assessment purposes.

# BMI CONSTANTS
# Again, John Brower Minnoch comes to our rescue here. HIs body mass index
# at the time of his death was 105 kg/m2. In terms of the lowest body mass index ever recorded
# there was nly limited web search information that suggested 7.5 kg.m2 as the lowest ever.

# GESTATION CONSTANTS
MINIMUM_GESTATION_WEEKS = 22
MINIMUM_GESTATION_WEEKS_ERROR = f"It is very unlikely that a gestational age of less than {MINIMUM_GESTATION_WEEKS} weeks is correct"
MAXIMUM_GESTATION_WEEKS = 44
MAXIMUM_GESTATION_WEEKS_ERROR = f"It is very unlikely that a gestational age of more than {MAXIMUM_GESTATION_WEEKS} weeks is correct"

# PARENTAL HEIGHT CONSTANTS
MINIMUM_PARENTAL_HEIGHT_CM = 50
# the shortest man in the world was 54.6 cm Chandra Bahadur Dangi
# the shortest woman in the world is Jyoti Kishanji Amge at 62.8 cm
# lower limit to paternal and maternal height here therefore set arbitrarily at 50 cm
# (this constant was added from the API server `schemas/request_validation_classes.py`)

# These constants are used to determine the range of SDS values that are considered - see discussion in issue #32 in the rcpchgrowth-python repository
# All previous hard-coded values have been replaced with these constants
MINIMUM_HEIGHT_WEIGHT_OFC_ADVISORY_SDS = -4
MAXIMUM_HEIGHT_WEIGHT_OFC_ADVISORY_SDS = 4
MINIMUM_BMI_ADVISORY_SDS = -4
MAXIMUM_BMI_ADVISORY_SDS = 4
MINIMUM_HEIGHT_WEIGHT_OFC_ERROR_SDS = -8
MAXIMUM_HEIGHT_WEIGHT_OFC_ERROR_SDS = 8
MINIMUM_BMI_ERROR_SDS = -15
MAXIMUM_BMI_ERROR_SDS = 15



import pytest

from rcpchgrowth.constants import MALE, FEMALE, UK_WHO, CDC
from rcpchgrowth.mid_parental_height import mid_parental_height, mid_parental_height_z, expected_height_z_from_mid_parental_height_z

maternal_height = 151
paternal_height = 167
ACCURACY = 1e-3


def test_midparental_height():
    assert mid_parental_height(maternal_height=maternal_height, paternal_height=paternal_height, sex=MALE) == 165.5 
    assert mid_parental_height(maternal_height=maternal_height, paternal_height=paternal_height, sex=FEMALE) == 152.5
    assert mid_parental_height_z(maternal_height=maternal_height, paternal_height=paternal_height, reference=UK_WHO) == pytest.approx(-0.8943229, ACCURACY)
    assert mid_parental_height_z(maternal_height=maternal_height, paternal_height=paternal_height, reference=CDC) == pytest.approx(-0.8177165233046019, ACCURACY)
    assert expected_height_z_from_mid_parental_height_z(mid_parental_height_z=-0.8943229) == pytest.approx(-0.44716145, ACCURACY)


class BoneAge:

    """
    The BoneAge class returns a bone age object
    Params
    `bone_age`: an estimated skeletal age calculated from xrays reported in decimal years
    `bone_age_sds`: an SDS for the bone age, based on references
    `bone_age_centile`: a centile for the bone age, based on references
    `bone_age_reference`: enum ['greulich-pyle', 'tanner-whitehouse-ii', 'tanner-whitehouse-iii', 'fels', 'bonexpert']
    """

    def __init__(
        self,
        decimal_age: float,
        bone_age: float,
        height: float,
        bone_age_type: str = None,
        bone_age_sds: float = None,
        bone_age_centile: float = None,
        bone_age_text: str = None,
        ):

        self.decimal_age = decimal_age
        self.bone_age = bone_age
        self.height=height
        self.bone_age_type = bone_age_type
        self.bone_age_sds = bone_age_sds
        self.bone_age_centile = bone_age_centile
        self.bone_age_text = bone_age_text

        return {
            "decimal_age": self.decimal_age,
            "bone_age": self.bone_age,
            "bone_age_type": self.bone_age_type,
            "bone_age_sds": self.bone_age_sds,
            "bone_age_centile": self.bone_age_centile,
            "bone_age_text": self.bone_age_text,
            "plottable_data": [
                {
                    "x": self.decimal_age,
                    "y": self.height
                },
                {
                    "x": self.decimal_age,
                    "y": self.bone_age
                }
            ]
        }


from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="rcpchgrowth",
    version="4.3.1",
    description="SDS and Centile calculations for UK Growth Data",
    long_description=long_description,
    url="https://github.com/rcpch/digital-growth-charts/blob/master/README.md",
    author="@eatyourpeas, @marcusbaw, @statist7, RCPCH Incubator",
    author_email="incubator@rcpch.ac.uk",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
    ],
    keywords="growth charts, anthropometry, SDS, centile, UK-WHO, UK90, Trisomy 21 (UK), Trisomy 21 (AAP), Turner, CDC",
    packages=find_packages(),
    python_requires=">3.8",
    install_requires=["python-dateutil", "scipy"],
    include_package_data=True,
    project_urls={
        "Bug Reports": "https://github.com/rcpch/rcpchgrowth-python/issues",
        "API management": "https://dev.rcpch.ac.uk",
        "Source": "https://github.com/rcpch/rcpchgrowth-python",
    },
)



from .constants import HEIGHT, MALE, FEMALE, UK_WHO, WHO
from .global_functions import sds_for_measurement
"""
Functions to calculate mid-parental height

cf 
Tanner JM, Whitehouse RH, Takaishi M. Standards from birth to maturity for height, weight, height velocity, and weight velocity: British children, 1965. I. Arch Dis Child. 1966;41(219):454-471.
The strengths and limitations of parental heights as a predictor of attained height, Charlotte M Wright, Tim D Cheetham, Arch Dis Child 1999;81:257–260
"""

def mid_parental_height(maternal_height, paternal_height, sex):
    """
    Calculate mid-parental height
    """
    if sex == MALE:
        return (maternal_height + paternal_height + 13) / 2
    else:
        return (maternal_height + paternal_height - 13) / 2

def mid_parental_height_z(maternal_height, paternal_height, reference=UK_WHO):
    """
    Calculate mid-parental height standard deviation
    """
    
    # convert parental heights to z-scores
    adult_age = 20.0
    if reference == WHO:
        adult_age = 19.0
    
    maternal_height_z = sds_for_measurement(reference=reference, age=adult_age, measurement_method=HEIGHT, observation_value=maternal_height, sex=FEMALE)
    paternal_height_z = sds_for_measurement(reference=reference, age=adult_age, measurement_method=HEIGHT, observation_value=paternal_height, sex=MALE)

    # take the means of the z-scores and apply the regression coefficient of 0.5 - simplifed: (MatHtz +PatHtz)/4
    mid_parental_height_z_score = (maternal_height_z + paternal_height_z) / 4.0

    return mid_parental_height_z_score

def expected_height_z_from_mid_parental_height_z(mid_parental_height_z):
    """
    Calculate expected height z score from mid-parental height z-score

    Ninety per cent of children had values within 1.4 SDS of their expected SDS (just over two
    centile spaces) and only 1% had values > 2 SDS (three centile spaces) below (cf Wright et al)
    """
    
    return mid_parental_height_z * 0.5

def lower_and_upper_limits_of_expected_height_z(mid_parental_height_z):
    """
    Calculate lower and upper limits of expected height z score from mid-parental height z-score
    Returns a tuple of (lower, upper) limits
    """
    
    return mid_parental_height_z - 1.4, mid_parental_height_z + 1.4


import json
from importlib import resources
from pathlib import Path
from .constants import *
# from .global_functions import z_score, cubic_interpolation, linear_interpolation, centile, measurement_for_z, nearest_lowest_index, fetch_lms
# import timeit #see below, comment back in if timing functions in this module

"""
birth_date: date of birth
observation_date: date of observation
sex: sex (string, MALE or FEMALE)
decimal_age: chronological, decimal
corrected_age: corrected for prematurity, decimal
measurement_method: height, weight, bmi, ofc (decimal)
observation: value (float)
gestation_weeks: gestational age(weeks), integer
gestation_days: supplementary days of gestation
lms: L, M or S
reference: reference data
"""

# load the reference data
data_directory = resources.files("rcpchgrowth.data_tables")



data_path = Path(data_directory, "trisomy_21_aap_infants.json")
with open(data_path) as json_file:
            TRISOMY_21_AAP_INFANT_DATA = json.load(json_file)
            json_file.close()

data_path = Path(data_directory, "trisomy_21_aap_children.json")
with open(data_path) as json_file:
            TRISOMY_21_AAP_CHILD_DATA = json.load(json_file)
            json_file.close()

def reference_data_absent( 
        age: float,
        measurement_method: str,
        sex: str
    ):
    """
    Helper function.
    Returns boolean
    Tests presence of valid reference data for a given measurement request

    Reference data is not complete for all ages/sexes/measurements.
     - There is only BMI reference data until 20y
     - Head circumference reference data is available until 18.0y
     - lowest threshold is 1 month, upper threshold is 20y
    """

    if age < 0: # lower threshold of trisomy_21 data
        return True, "No Trisomy 21 (AAP) reference data exists below 40 weeks gestation"
    
    if age < 0.083 and measurement_method in ['height', 'ofc']: # lower threshold of trisomy_21 data
        return True, "No Trisomy 21 (AAP) reference data exists below 1 month of age for height or head circumference."
    
    if age < 2 and measurement_method == 'bmi': # lower threshold of trisomy_21 data
         return True, "No Trisomy 21 (AAP) reference data exists below 2 years of age for BMI."
    
    if age > TWENTY_YEARS: # upper threshold of trisomy_21 data
        return True, "Trisomy 21 reference data does not exist over the age of 20y."
    
    return False, None

def trisomy_21_aap_lms_array_for_measurement_and_sex(
        measurement_method: str,
        sex: str,
        age: float,
        default_youngest_reference: bool = False
    ):
    # returns the LMS array for a given measurement
    # raises a LookupError if the data is not available
    # No need for default_youngest_reference as all the data is from the same source
    # Note there is an overlap in the age ranges of the two datasets, so the age parameter is used to select the correct dataset - below 36mths the data is more granular

    data_invalid, data_error = reference_data_absent(age=age, measurement_method=measurement_method, sex=sex)

    if data_invalid:
        raise LookupError(f"Reference data absent: {data_error}")
    else:
        if age <= 3.0 and measurement_method in ['height', 'weight', 'ofc']:
            if measurement_method == 'ofc' and age==3 and default_youngest_reference:
                return TRISOMY_21_AAP_INFANT_DATA["measurement"][measurement_method][sex]
            elif measurement_method == 'ofc' and age==3 and not default_youngest_reference:
                return TRISOMY_21_AAP_CHILD_DATA["measurement"][measurement_method][sex]
            
            return TRISOMY_21_AAP_INFANT_DATA["measurement"][measurement_method][sex]
        else:
            return TRISOMY_21_AAP_CHILD_DATA["measurement"][measurement_method][sex]

def select_reference_data_for_trisomy_21_aap(trisomy_21_aap_reference_name, measurement_method:str, sex:str, default_youngest_reference: bool = False):

    if trisomy_21_aap_reference_name == TRISOMY_21_AAP_INFANT:
        try:
            return_value = trisomy_21_aap_lms_array_for_measurement_and_sex(measurement_method=measurement_method, sex=sex, age=1.0, default_youngest_reference=default_youngest_reference) # select arbitrary age of 1 y for infant data
        except:
            raise LookupError(f"No data for {measurement_method} in the {sex} Trisomy 21 (AAP) dataset (<36mths).")
    elif trisomy_21_aap_reference_name == TRISOMY_21_AAP_CHILD:
        try:
            return_value = trisomy_21_aap_lms_array_for_measurement_and_sex(measurement_method=measurement_method, sex=sex, age=4.0, default_youngest_reference=default_youngest_reference) # select arbitrary age of 4 y for child data
        except:
            raise LookupError(f"No data for {measurement_method} in the {sex} Trisomy 21 (US) dataset (>3y).")
    return return_value